/*
 * Copyright (c) 2015 N. Billard.
 */
package com.massa.mrules.demos.classroom;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import com.massa.log.Log;
import com.massa.log.LogFactory;
import com.massa.mrules.context.compile.CompilationLevel;
import com.massa.mrules.context.execute.MExecutionContext;
import com.massa.mrules.demos.classroom.model.ClassRoom;
import com.massa.mrules.demos.classroom.model.Mark;
import com.massa.mrules.demos.classroom.model.MentionType;
import com.massa.mrules.demos.classroom.model.Student;
import com.massa.mrules.exception.MConfigurationException;
import com.massa.mrules.exception.MExecutionException;
import com.massa.mrules.factory.ser.SerializationFactory;
import com.massa.mrules.factory.xml.RichXMLFactory;
import com.massa.mrules.set.IMruleExecutionSet;

@RunWith(Parameterized.class)
public class ClassRoomTest {
	private static final Log LOG = LogFactory.getLog(ClassRoomTest.class.getName());

	final CompilationLevel level;
	final String xmlName;

	public ClassRoomTest(final CompilationLevel level, final String xmlName) {
		this.level = level;
		this.xmlName = xmlName;
	}

	@Parameters(name="{0} - {1}")
	public static Collection<Object[]> configs() {
		return Arrays.asList(new Object[][] {
			{CompilationLevel.STANDARD, ClassRoomTest.XML1},
			{CompilationLevel.STANDARD, ClassRoomTest.XML2},
			{CompilationLevel.LIGHT, ClassRoomTest.XML1},
			{CompilationLevel.LIGHT, ClassRoomTest.XML2}
		});
	}


	private static final Integer[] COEF = new Integer[] {
			1,     1,     2,     1,     1,     3,     2,     4
	};
	private static final Double [][] MARKS = new Double[][] {
		{12.5D, 15D,   8D,    10.5D, 13D,   7D,    9D,    11D},
		{4D,    3D,    6.5D,  9D,    4D,    2D,    2D,    7D},
		{18D,   13D,   14D,   15D,   16D,   17D,   11D,   17D},
		{10D,   8D,    12D,   13D,   14D,   12.5D, 11.5D, 13D},
		{9D,    13D,   8D,    9.5D,  8.5D,  9.5D,  7.5D,  8D},
		{18D,   13D,   14D,   15D,   16D,   17D,   null,   17D},
		{11D,   14D,   9D,    3D,    17D,   null,  12D,   11.5D},
	};

	public static final String XML1 = "com/massa/mrules/demos/classroom/rules_1.xml";
	public static final String XML2 = "com/massa/mrules/demos/classroom/rules_2.xml";

	@Test
	public void classRoomDemo() throws MConfigurationException, MExecutionException, IOException {
		final ClassRoom classroomMan = ClassRoomTest.buildClassRoom();
		final ClassRoom classroomEng = ClassRoomTest.buildClassRoom();

		final RichXMLFactory factory = new RichXMLFactory(this.level);

		final InputStream xml = ClassLoader.getSystemResourceAsStream(this.xmlName);
		final IMruleExecutionSet eng = factory.read(xml);

		@SuppressWarnings("unchecked")
		final
		MExecutionContext<ClassRoom> ctxt = (MExecutionContext<ClassRoom>) eng.getContextFactory().getExecutionContext();
		ctxt.setInput(classroomEng);
		ctxt.execute();

		ClassRoomTest.manualCompute(classroomMan);

		ClassRoomTest.compare(classroomMan, classroomEng);

		//Time check
		final ArrayList<ClassRoom> l = new ArrayList<ClassRoom>();
		for (int i = 0 ; i < 10000 ; i++) {
			l.add(ClassRoomTest.buildClassRoom());
		}

		ClassRoomTest.LOG.info("Beginning performance test.");
		final long t1 = System.currentTimeMillis();
		ctxt.setOut(null); //To avoid thousands of printings.
		for (final ClassRoom classroom: l) {
			ctxt.setInput(classroom);
			ctxt.execute();
		}
		ClassRoomTest.LOG.info("Time for 10000 tests : " + (System.currentTimeMillis() - t1));

		//Read / Write tests.
		ClassRoomTest.LOG.info("Human readable version:\n" + eng.toString());

		final StringWriter w = new StringWriter();
		factory.write(w, eng);
		ClassRoomTest.LOG.info(w.toString());

		final ByteArrayOutputStream o = new ByteArrayOutputStream();
		final SerializationFactory sfactory = new SerializationFactory(this.level);
		sfactory.write(o, eng);

		Assert.assertEquals(eng, eng.clone());
		Assert.assertEquals(eng, factory.read(new StringReader(w.toString())));
		Assert.assertEquals(eng, sfactory.read(new ByteArrayInputStream(o.toByteArray())));
	}

	public static void compare(final ClassRoom classroomMan, final ClassRoom classroomEng) {
		for (int iStudent = 0 ; iStudent < classroomMan.getStudents().size() ; iStudent++) {
			Assert.assertEquals("index: " + iStudent,
					classroomMan.getStudents().get(iStudent).getAverage(),
					classroomEng.getStudents().get(iStudent).getAverage(),
					0.00001);
			Assert.assertEquals("index: " + iStudent,
					classroomMan.getStudents().get(iStudent).isAdmitted(),
					classroomEng.getStudents().get(iStudent).isAdmitted());
			Assert.assertEquals("index: " + iStudent,
					classroomMan.getStudents().get(iStudent).isMayRepass(),
					classroomEng.getStudents().get(iStudent).isMayRepass());
			Assert.assertEquals("index: " + iStudent,
					classroomMan.getStudents().get(iStudent).getMention(),
					classroomEng.getStudents().get(iStudent).getMention());
		}
		Assert.assertEquals(classroomMan.getClassAverage(), classroomEng.getClassAverage(), 0.00001);
		Assert.assertEquals(classroomMan.getNumberOfAdmitted(), classroomEng.getNumberOfAdmitted());
		Assert.assertEquals(classroomMan.getNumberOfRejected(), classroomEng.getNumberOfRejected());


	}

	public static void manualCompute(final ClassRoom classroom) {
		double sumAverage = 0;
		for (final Student student: classroom.getStudents()) {
			int sumCoef = 0;
			double sumMarks = 0;
			for (int iMark = 0 ; iMark < student.getMarks().size() ; iMark++) {
				if (student.getMarks().get(iMark) != null) {
					sumMarks += student.getMarks().get(iMark).getMark() * student.getMarks().get(iMark).getCoefficient();
					sumCoef += classroom.getCoefficients().get(iMark);
				}
			}
			final double average = Math.round(10 * sumMarks / sumCoef) / 10.0;
			sumAverage += average;
			student.setAverage(average);
			student.setAdmitted(average >= 10);
			student.setMayRepass(average < 10 && average >= 8);
			student.setMention(average >= 16 ? MentionType.GREAT : average >= 14 ? MentionType.VERY_GOOD : average > 12 ? MentionType.GOOD : null);
			if (average >= 10) {
				classroom.setNumberOfAdmitted(classroom.getNumberOfAdmitted() + 1);
			} else {
				classroom.setNumberOfRejected(classroom.getNumberOfRejected() + 1);
			}
		}
		classroom.setClassAverage(Math.round(100 * sumAverage / classroom.getStudents().size()) / 100.0);
	}



	public static ClassRoom buildClassRoom() {
		final ClassRoom ret = new ClassRoom();
		ret.setCoefficients(Arrays.asList(ClassRoomTest.COEF));
		ret.setStudents(new ArrayList<Student>());
		for (int iStudent = 0 ; iStudent < ClassRoomTest.MARKS.length ; iStudent++) {
			final Student s = new Student();
			s.setName("Student " + iStudent);
			s.setMarks(new ArrayList<Mark>());
			for (int iMark = 0 ; iMark < ClassRoomTest.MARKS[iStudent].length ; iMark++) {
				Mark m = null;
				if (ClassRoomTest.MARKS[iStudent][iMark] != null) {
					m = new Mark();
					m.setMark(ClassRoomTest.MARKS[iStudent][iMark]);
					m.setCoefficient(ClassRoomTest.COEF[iMark]);
				}
				s.getMarks().add(m);
			}
			ret.getStudents().add(s);
		}
		return ret;
	}
}
